@@ -1,5 +1,6 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE OverloadedLabels #-}
+{-# LANGUAGE CPP #-}
 module StatusNotifier.Tray where
 
 import           Control.Concurrent.MVar as MV
@@ -11,11 +12,13 @@ import           DBus.Client
 import qualified DBus.Internal.Types as DBusTypes
 import qualified Data.ByteString as BS
 import           Data.Coerce
+import           Data.Foldable (traverse_)
 import           Data.Int
 import           Data.List
 import qualified Data.Map.Strict as Map
 import           Data.Maybe
 import           Data.Ord
+import           Data.Ratio
 import qualified Data.Text as T
 import qualified GI.DbusmenuGtk3.Objects.Menu as DM
 import qualified GI.GLib as GLib
@@ -23,6 +26,7 @@ import           GI.GLib.Structs.Bytes
 import qualified GI.Gdk as Gdk
 import           GI.Gdk.Enums
 import           GI.Gdk.Objects.Screen
+import           GI.Gdk.Structs.EventScroll
 import           GI.GdkPixbuf.Enums
 import           GI.GdkPixbuf.Objects.Pixbuf
 import qualified GI.Gtk as Gtk
@@ -112,7 +116,8 @@ getIconPixbufByName size name themePath = do
   then do
     let targetName = if hasPanelIcon then panelName else name
     trayLogger DEBUG $ printf "Found icon %s in theme" name
-    iconThemeLoadIcon themeForIcon targetName size themeLoadFlags
+    catchAny (iconThemeLoadIcon themeForIcon targetName size themeLoadFlags)
+             (const $ pure Nothing)
 
   else do
     trayLogger DEBUG $ printf "Trying to load icon %s as filepath" name
@@ -122,7 +127,12 @@ getIconPixbufByName size name themePath = do
     maybeFile <- if fileExists
     then return $ Just nameString
     else fmap join $ sequenceA $ getIconPathFromThemePath nameString <$> themePath
-    sequenceA $ pixbufNewFromFile <$> maybeFile
+#if MIN_VERSION_gi_gdkpixbuf(2,0,26)
+    let handleResult = fmap join . sequenceA
+#else
+    let handleResult = sequenceA
+#endif
+    handleResult $ pixbufNewFromFile <$> maybeFile
 
 getIconPathFromThemePath :: String -> String -> IO (Maybe String)
 getIconPathFromThemePath name themePath = if name == "" then return Nothing else do
@@ -156,28 +166,45 @@ data ItemContext = ItemContext
 
 data TrayImageSize = Expand | TrayImageSize Int32
 
+data TrayClickAction = Activate | SecondaryActivate | PopupMenu
+
 data TrayParams = TrayParams
-  { trayHost :: Host
-  , trayClient :: Client
-  , trayOrientation :: Gtk.Orientation
+  { trayOrientation :: Gtk.Orientation
   , trayImageSize :: TrayImageSize
   , trayIconExpand :: Bool
   , trayAlignment :: StrutAlignment
   , trayOverlayScale :: Rational
+  , trayLeftClickAction :: TrayClickAction
+  , trayMiddleClickAction :: TrayClickAction
+  , trayRightClickAction :: TrayClickAction
   }
 
-buildTray :: TrayParams -> IO Gtk.Box
-buildTray TrayParams { trayHost = Host
-                       { itemInfoMap = getInfoMap
-                       , addUpdateHandler = addUHandler
-                       , removeUpdateHandler = removeUHandler
-                       }
-                     , trayClient = client
-                     , trayOrientation = orientation
+defaultTrayParams = TrayParams
+  { trayOrientation = Gtk.OrientationHorizontal
+  , trayImageSize = Expand
+  , trayIconExpand = False
+  , trayAlignment = End
+  , trayOverlayScale = 3 % 5
+  , trayLeftClickAction = Activate
+  , trayMiddleClickAction = SecondaryActivate
+  , trayRightClickAction = PopupMenu
+  }
+
+buildTray :: Host -> Client -> TrayParams -> IO Gtk.Box
+buildTray Host
+            { itemInfoMap = getInfoMap
+            , addUpdateHandler = addUHandler
+            , removeUpdateHandler = removeUHandler
+            }
+          client
+          TrayParams { trayOrientation = orientation
                      , trayImageSize = imageSize
                      , trayIconExpand = shouldExpand
                      , trayAlignment = alignment
                      , trayOverlayScale = overlayScale
+                     , trayLeftClickAction = leftClickAction
+                     , trayMiddleClickAction = middleClickAction
+                     , trayRightClickAction = rightClickAction
                      } = do
   trayLogger INFO "Building tray"
 
@@ -194,7 +221,10 @@ buildTray TrayParams { trayHost = Host
           _ ->
             Gdk.getRectangleWidth rectangle
 
-      getInfo def name = fromMaybe def . Map.lookup name <$> getInfoMap
+      getInfoAttr fn def name = maybe def fn . Map.lookup name <$> getInfoMap
+
+      getInfo :: ItemInfo -> DBusTypes.BusName -> IO ItemInfo
+      getInfo = getInfoAttr id
 
       updateIconFromInfo info@ItemInfo { itemServiceName = name } =
         getContext name >>= updateIcon
@@ -237,6 +267,7 @@ buildTray TrayParams { trayHost = Host
           trayLogger INFO logText
 
           button <- Gtk.eventBoxNew
+          Gtk.widgetAddEvents button [Gdk.EventMaskScrollMask]
 
           image <-
             case imageSize of
@@ -256,8 +287,9 @@ buildTray TrayParams { trayHost = Host
                           in return (thisTime, thisTime /= previous)
 
                         trayLogger DEBUG $
-                                   printf "Allocating image size %s, width %s, \
-                                           \ height %s, resize %s"
+                                   printf
+                                   ("Allocating image size %s, width %s," <>
+                                    " height %s, resize %s")
                                    (show size)
                                    (show actualWidth)
                                    (show actualHeight)
@@ -304,11 +336,41 @@ buildTray TrayParams { trayHost = Host
               popupItemForMenu menu =
                 Gtk.menuPopupAtWidget menu image
                    GravitySouthWest GravityNorthWest Nothing
-              popupItemMenu =
-                maybe activateItem popupItemForMenu maybeMenu >> return False
-              activateItem = void $ IC.activate client serviceName servicePath 0 0
 
-          _ <- Gtk.onWidgetButtonPressEvent button $ const popupItemMenu
+          _ <- Gtk.onWidgetButtonPressEvent button $ \event -> do
+            button <- Gdk.getEventButtonButton event
+            x <- round <$> Gdk.getEventButtonXRoot event
+            y <- round <$> Gdk.getEventButtonYRoot event
+            isMenu <- getInfoAttr itemIsMenu False serviceName
+            let action = if isMenu then PopupMenu else
+                  case button of
+                    1 -> leftClickAction
+                    2 -> middleClickAction
+                    3 -> rightClickAction
+            case action of
+              Activate -> void $ IC.activate client serviceName servicePath x y
+              SecondaryActivate -> void $ IC.secondaryActivate client
+                                   serviceName servicePath x y
+              PopupMenu -> maybe (return ()) popupItemForMenu maybeMenu
+            return False
+          _ <- Gtk.onWidgetScrollEvent button $ \event -> do
+            direction <- getEventScrollDirection event
+            let direction' = case direction of
+                               ScrollDirectionUp -> Just "vertical"
+                               ScrollDirectionDown -> Just "vertical"
+                               ScrollDirectionLeft -> Just "horizontal"
+                               ScrollDirectionRight -> Just "horizontal"
+                               _ -> Nothing
+                -- deltaX/deltaY are provided only in case of smooth scrolling which
+                -- is enabled via additional flag, we don't to enable/handle it
+                delta = case direction of
+                          ScrollDirectionUp -> -1
+                          ScrollDirectionDown -> 1
+                          ScrollDirectionLeft -> -1
+                          ScrollDirectionRight -> 1
+                          _ -> 0
+            traverse_ (IC.scroll client serviceName servicePath delta) direction'
+            return False
 
           MV.modifyMVar_ contextMap $ return . Map.insert serviceName context
 
@@ -369,10 +431,12 @@ buildTray TrayParams { trayHost = Host
                                       } = getPixBufFrom size name mpath pixmaps
 
       getOverlayPixBufFromInfo size
-                               info@ItemInfo { overlayIconName = name
-                                             , iconThemePath = mpath
-                                             , overlayIconPixmaps = pixmaps
-                                             } = getPixBufFrom size (fromMaybe "" name) mpath pixmaps
+                               info@ItemInfo
+                                     { overlayIconName = name
+                                     , iconThemePath = mpath
+                                     , overlayIconPixmaps = pixmaps
+                                     } = getPixBufFrom size (fromMaybe "" name)
+                               mpath pixmaps
 
       getPixBufFrom size name mpath pixmaps = do
         let tooSmall (w, h, _) = w < size || h < size
